#include <msp/builder/component.h>
#include <msp/builder/sourcepackage.h>
#include <msp/core/algorithm.h>
#include <msp/fs/utils.h>
#include "androidassetpackagingtool.h"
#include "androidmanifestfile.h"
#include "androidresourcebundle.h"
#include "androidtools.h"

using namespace std;
using namespace Msp;

AndroidAssetPackagingTool::AndroidAssetPackagingTool(Builder &b, const AndroidSdk &s):
	Tool(b, "AAPT"),
	sdk(s)
{
	if(sdk.get_root_dir().empty())
		problems.push_back("Android SDK not found");
	else if(sdk.get_build_tools_dir().empty())
		problems.push_back("Android build-tools not found");
	else
		set_command((sdk.get_build_tools_dir()/"aapt").str());

	if(sdk.get_platform_jar().empty())
		problems.push_back("Android platform not found");

	set_run_external(_run);
}

Target *AndroidAssetPackagingTool::create_target(const vector<Target *> &sources, const string &)
{
	AndroidManifestFile *manifest = 0;
	vector<FileTarget *> resources;
	resources.reserve(sources.size());
	for(Target *s: sources)
	{
		if(AndroidManifestFile *m = dynamic_cast<AndroidManifestFile *>(s))
			manifest = m;
		else if(FileTarget *f = dynamic_cast<FileTarget *>(s))
			resources.push_back(f);
	}

	if(!manifest)
		throw invalid_argument("AndroidAssetPackagingTool::create_target");

	AndroidResourceBundle *res = new AndroidResourceBundle(builder, *manifest->get_component(), *manifest, resources);
	res->set_tool(*this);
	return res;
}

ExternalTask::Arguments AndroidAssetPackagingTool::_run(const AndroidResourceBundle &res, FS::Path &work_dir)
{
	const AndroidAssetPackagingTool &tool = dynamic_cast<const AndroidAssetPackagingTool &>(*res.get_tool());

	ExternalTask::Arguments argv;
	argv.push_back(tool.get_executable()->get_path().str());
	argv.push_back("package");

	argv.push_back("-I");
	argv.push_back(tool.sdk.get_platform_jar().str());

	argv.push_back("-F");
	argv.push_back(FS::relative(res.get_path(), work_dir).str());

	vector<FS::Path> resource_dirs;
	resource_dirs.reserve(res.get_dependencies().size());
	for(Target *d: res.get_dependencies())
	{
		FileTarget *file = dynamic_cast<FileTarget *>(d);
		Target *real = d->get_real_target();

		if(dynamic_cast<AndroidManifestFile *>(real))
		{
			argv.push_back("-M");
			argv.push_back(FS::relative(file->get_path(), work_dir).str());
		}
		else if(real->get_package()==res.get_package())
		{
			const FS::Path &path = file->get_path();
			FS::Path res_dir = path.subpath(0, path.size()-2);
			if(!any_equals(resource_dirs, res_dir))
				resource_dirs.push_back(res_dir);
		}
	}

	for(const FS::Path &d: resource_dirs)
	{
		argv.push_back("-S");
		argv.push_back(FS::relative(d, work_dir).str());
	}

	return argv;
}
