#include <msp/builder/builder.h>
#include <msp/builder/component.h>
#include <msp/builder/csourcefile.h>
#include <msp/builder/objectfile.h>
#include <msp/builder/sourcepackage.h>
#include <msp/core/environ.h>
#include <msp/core/maputils.h>
#include <msp/fs/utils.h>
#include <msp/strings/format.h>
#include <msp/strings/utils.h>
#include "microsofttools.h"
#include "msvccompiler.h"

using namespace std;
using namespace Msp;

MsvcCompiler::MsvcCompiler(Builder &b, const Architecture &a, const string &t, const MicrosoftTools &m):
	Tool(b, &a, t),
	ms_tools(m)
{
	if(tag=="CC")
	{
		input_suffixes.push_back(".c");
		aux_suffixes.push_back(".h");
	}
	else if(tag=="CXX")
	{
		input_suffixes.push_back(".cpp");
		input_suffixes.push_back(".cc");
		aux_suffixes.push_back(".hpp");
	}
	else
		throw invalid_argument("MsvcCompiler::MsvcCompiler");

	set_command((ms_tools.get_vc_bin_dir()/"cl.exe").str(), false);
	set_run_external(_run);
}

Target *MsvcCompiler::create_source(const Component &comp, const FS::Path &path) const
{
	return new CSourceFile(builder, comp, path);
}

Target *MsvcCompiler::create_source(const FS::Path &path) const
{
	return new CSourceFile(builder, path);
}

Target *MsvcCompiler::create_target(const vector<Target *> &sources, const string &)
{
	if(sources.size()!=1)
		throw invalid_argument("MsvcCompiler::create_target");
	SourceFile &source = dynamic_cast<SourceFile &>(*sources.front());
	ObjectFile *obj = new ObjectFile(builder, *source.get_component(), source);
	obj->set_tool(*this);
	return obj;
}

string MsvcCompiler::create_build_signature(const BuildInfo &binfo) const
{
	string result = Tool::create_build_signature(binfo);
	result += ',';
	if(binfo.debug)
		result += 'g';
	if(binfo.optimize)
	{
		result += 'O';
		result += (binfo.optimize<0 ? '1' : binfo.optimize==1 ? 'x' : '2');
	}
	if(binfo.libmode<=BuildInfo::STATIC)
		result += 't';
	return result;
}

void MsvcCompiler::do_prepare(ToolData &tool) const
{
	const std::string &tool_tag = static_cast<Tool &>(tool).get_tag();

	const FS::Path &vc_base_dir = ms_tools.get_vc_base_dir();
	tool.system_path.push_back(vc_base_dir/"include");

	const FS::Path &win_sdk_dir = ms_tools.get_windows_sdk_dir();
	const string &win_sdk_ver = ms_tools.get_windows_sdk_version();
	tool.system_path.push_back(win_sdk_dir/"include"/win_sdk_ver/"ucrt");
	tool.system_path.push_back(win_sdk_dir/"include"/win_sdk_ver/"shared");
	tool.system_path.push_back(win_sdk_dir/"include"/win_sdk_ver/"um");

	string path;
	for(const FS::Path &p: tool.system_path)
	{
		append(path, ";", p.str());
		builder.get_logger().log("tools", "Got %s system path: %s", tool_tag, p);
	}

	setenv("INCLUDE", path);
}

ExternalTask::Arguments MsvcCompiler::_run(const ObjectFile &object, FS::Path &work_dir)
{
	const Tool &tool = *object.get_tool();

	ExternalTask::Arguments argv;
	argv.push_back(tool.get_executable()->get_path().str());
	argv.push_back("/nologo");
	argv.push_back("/c");

	BuildInfo binfo;
	object.collect_build_info(binfo);

	if(binfo.standards.count(tool.get_tag()))
	{
		const BuildInfo::LanguageStandard &std = get_item(binfo.standards, tool.get_tag());
		if((tool.get_tag()=="CXX" && std.year>2011) || (tool.get_tag()=="CC" && std.year>1999))
			argv.push_back("/std:"+std.str());
	}

	if(binfo.warning_level>=1)
	{
		argv.push_back(format("/W%d", binfo.warning_level));
		if(binfo.fatal_warnings)
			argv.push_back("/WX");
		if(binfo.warning_level>=3)
			argv.push_back("/permissive-");

		argv.push_back("/wd4068");  // Unknown pragma
		argv.push_back("/wd4251");  // Dll-interface required on type of member variable
		argv.push_back("/wd4275");  // Dll-interface class derives from non-dll-interface class
		if(binfo.warning_level<4)
		{
			argv.push_back("/wd4244");  // Narrowing conversion on arg or return
			argv.push_back("/wd4267");  // Narrowing conversion
		}
	}
	else
		argv.push_back("/w");

	for(const FS::Path &p: binfo.local_incpath)
	{
		argv.push_back("/I");
		argv.push_back(p.str());
	}
	for(const FS::Path &p: binfo.incpath)
	{
		argv.push_back("/I");
		argv.push_back(p.str());
	}

	for(const auto &kvp: binfo.defines)
	{
		argv.push_back("/D");
		if(kvp.second.empty())
			argv.push_back(kvp.first);
		else
			argv.push_back(format("%s=%s", kvp.first, kvp.second));
	}

	if(binfo.debug)
		argv.push_back("/Z7");
	if(binfo.optimize)
	{
		if(binfo.optimize<0)
			argv.push_back("/O1");
		else if(binfo.optimize==1)
			argv.push_back("/Ox");
		else
			argv.push_back("/O2");
	}

	if(binfo.libmode<=BuildInfo::STATIC)
		argv.push_back(binfo.debug ? "/MTd" : "/MT");
	else
		argv.push_back(binfo.debug ? "/MDd" : "/MD");

	argv.push_back("/EHsc");

	FS::Path obj_path = object.get_path();
	FS::Path src_path = object.get_source().get_path();

	argv.push_back("/Fo"+relative(obj_path, work_dir).str());
	argv.push_back(relative(src_path, work_dir).str());

	return argv;
}
