#ifndef ARCHITECTURE_H_
#define ARCHITECTURE_H_

#include <typeinfo>
#include <msp/datafile/loader.h>
#include "buildinfo.h"
#include "libbuilder_api.h"
#include "pattern.h"

class Builder;

/**
Stores information about an architecture.  This includes CPU type, model and
bitness and operating system.
*/
class LIBBUILDER_API Architecture
{
public:
	class Loader: public Msp::DataFile::ObjectLoader<Architecture>
	{
	public:
		Loader(Architecture &);

	private:
		void cross_prefix(const std::string &);
	};

private:
	Builder &builder;
	std::string type;
	std::string cpu;
	std::string fpu;
	std::string system;
	unsigned bits = 0;
	std::string toolchain;
	std::string name;
	bool native = false;
	std::string cross_prefix;
	std::map<std::string, std::vector<Pattern>> filename_patterns;

public:
	Architecture(Builder &b, const std::string &spec);

	void refine(const std::string &);
private:
	void update();

public:
	const std::string &get_type() const { return type; }
	const std::string &get_name() const { return name; }
	const std::string &get_system() const { return system; }
	unsigned get_bits() const { return bits; }
	const std::string &get_cpu() const { return cpu; }
	const std::string &get_fpu() const { return fpu; }
	const std::string &get_toolchain() const { return toolchain; }
	bool match_name(const std::string &, unsigned * = 0) const;
	std::string best_match(const std::vector<std::string> &) const;
	bool is_native() const { return native; }
	bool is_cross() const { return !cross_prefix.empty(); }

	const std::string &get_cross_prefix() const { return cross_prefix; }

	template<typename T>
	const std::vector<Pattern> &get_patterns() const;

	template<typename T>
	std::string create_filename(const std::string &) const;

private:
	template<typename T>
	void add_pattern(const std::string &);

private:
	static void resolve_aliases(std::vector<std::string> &);
	void parse_specification(const std::string &);
};

template<typename T>
inline const std::vector<Pattern> &Architecture::get_patterns() const
{
	auto i = filename_patterns.find(typeid(T).name());
	if(i!=filename_patterns.end())
		return i->second;

	static std::vector<Pattern> empty;
	return empty;
}

template<typename T>
inline std::string Architecture::create_filename(const std::string &base) const
{
	const std::vector<Pattern> &patterns = get_patterns<T>();
	return patterns.empty() ? base : patterns.front().apply(base);
}

#endif
