#include <msp/core/algorithm.h>
#include <msp/fs/utils.h>
#include <msp/strings/format.h>
#include <msp/strings/utils.h>
#include "binary.h"
#include "builder.h"
#include "component.h"
#include "objectfile.h"
#include "sharedlibrary.h"
#include "sourcepackage.h"
#include "staticlibrary.h"
#include "tool.h"

using namespace std;
using namespace Msp;

Binary::Binary(Builder &b, const Component &c, const string &p, const vector<ObjectFile *> &objs):
	FileTarget(b, c.get_package(), c.get_package().get_output_directory()/p),
	objects(objs)
{
	component = &c;
	for(ObjectFile *o: objects)
		add_dependency(*o);

	nested_build_sig = true;
	arch_in_build_sig = true;
}

void Binary::collect_build_info(BuildInfo &binfo) const
{
	vector<const Tool *> tools;
	for(ObjectFile *o: objects)
		if(const Tool *obj_tool = o->get_tool())
			if(!any_equals(tools, obj_tool))
				tools.push_back(obj_tool);
	for(const Tool *t: tools)
		binfo.update_from(t->get_build_info());

	Target::collect_build_info(binfo);

	binfo.update_from(static_binfo);
}

void Binary::find_dependencies()
{
	if(!component)
		return;

	vector<Target *> static_libs;
	vector<Target *> shared_libs;
	vector<string> missing_libs;
	find_dependencies(this, static_libs, shared_libs, missing_libs);

	for(Target *t: static_libs)
		add_dependency(*t);
	for(Target *t: shared_libs)
		add_dependency(*t);
	for(const string &m: missing_libs)
		problems.push_back(format("Required library %s not found", m));
}

void Binary::find_dependencies(Target *tgt, vector<Target *> &static_libs, vector<Target *> &shared_libs, vector<string> &missing_libs)
{
	BuildInfo binfo;
	tgt->collect_build_info(binfo);
	if(tgt!=this)
	{
		static_binfo.libpath.insert(static_binfo.libpath.end(), binfo.libpath.begin(), binfo.libpath.end());
		static_binfo.keep_symbols.insert(static_binfo.keep_symbols.end(), binfo.keep_symbols.begin(), binfo.keep_symbols.end());
		if(binfo.threads)
			static_binfo.threads = true;
	}

	for(const string &l: binfo.libs)
	{
		if(l.size()>10 && !l.compare(l.size()-10, 10, ".framework"))
			continue;

		BuildInfo::LibraryMode libmode = binfo.get_libmode_for(l);
		Target *lib = builder.get_vfs().find_library(l, binfo.libpath, libmode);
		if(lib)
		{
			Target *real = lib->get_real_target();
			if(StaticLibrary *stlib = dynamic_cast<StaticLibrary *>(real))
			{
				/* Keep only the last occurrence of each static library.  This
				ensures the order is correct for linking. */
				auto i = find(static_libs, stlib);
				if(i!=static_libs.end())
					static_libs.erase(i);
				static_libs.push_back(stlib);
				find_dependencies(stlib, static_libs, shared_libs, missing_libs);
			}
			else if(!any_equals(shared_libs, lib))
				shared_libs.push_back(lib);
		}
		else if(!any_equals(missing_libs, l))
			missing_libs.push_back(l);
	}
}
