#include <msp/fs/utils.h>
#include "binarycomponent.h"
#include "builder.h"
#include "filetarget.h"
#include "sourcepackage.h"
#include "tool.h"

using namespace std;
using namespace Msp;

void BinaryComponent::create_build_info()
{
	Component::create_build_info();

	for(const Component *u: uses)
	{
		build_info.libs.push_back(u->get_name());
		if(!u->get_install())
		{
			build_info.libmodes[u->get_name()] = BuildInfo::STATIC;
			build_info.libpath.push_back(u->get_package().get_output_directory());
		}
	}

	if(type==LIBRARY || type==MODULE)
		if(build_info.libmode<BuildInfo::DYNAMIC)
			build_info.libmode = BuildInfo::DYNAMIC;
}

void BinaryComponent::update_exported_build_info(BuildInfo &binfo) const
{
	if(type==LIBRARY)
		binfo.libs.push_back(name);
}

void BinaryComponent::create_targets() const
{
	Builder &builder = package.get_builder();
	BuildGraph &build_graph = builder.get_build_graph();
	const Toolchain &toolchain = builder.get_toolchain();
	const Toolchain &pkg_tools = package.get_toolchain();

	Tool &linker = toolchain.get_tool("LINK");

	vector<FileTarget *> pending = create_sources();

	vector<Target *> objs;
	while(!pending.empty())
	{
		Tool *tool = 0;
		for(auto i=pending.begin(); i!=pending.end(); )
		{
			string ext = FS::extpart(FS::basename((*i)->get_path()));
			if(linker.accepts_suffix(ext))
			{
				/* Targets accepted by the linker should not be processed further
				before feeding them to the linker */
				objs.push_back(*i);
				i = pending.erase(i);
				continue;
			}

			// Pick the tool with the smallest processing unit
			Tool *tgt_tool = pkg_tools.get_tool_for_suffix(ext);
			if(!tgt_tool)
				tgt_tool = toolchain.get_tool_for_suffix(ext);
			if(tgt_tool && (!tool || tgt_tool->get_processing_unit()<tool->get_processing_unit()))
			{
				tool = tgt_tool;
				if(tool->get_processing_unit()==Tool::ONE_FILE)
					break;
			}

			++i;
		}

		if(!tool)
			break;

		Tool::ProcessingUnit processing_unit = tool->get_processing_unit();
		if(processing_unit==Tool::ONE_FILE)
		{
			// Process all pending accepted by the tool
			for(FileTarget *&t: pending)
			{
				string ext = FS::extpart(FS::basename(t->get_path()));
				if(tool->accepts_suffix(ext))
				{
					t = &dynamic_cast<FileTarget &>(*tool->create_target(*t));
					if(type==LIBRARY)
						create_install(*t);
				}
			}
		}
		else
		{
			// Process one group of pending accepted by the tool
			vector<Target *> group = extract_group(pending, *tool);
			FileTarget &tgt = dynamic_cast<FileTarget &>(*tool->create_target(group));
			pending.push_back(&tgt);
			if(type==LIBRARY)
				create_install(tgt);
		}
	}

	vector<Target *> results;
	results.reserve(3);
	if(type==LIBRARY)
	{
		Tool &archiver = toolchain.get_tool("AR");
		Target *shlib = linker.create_target(objs, "shared");
		results.push_back(shlib);
		if(Target *imp = linker.create_target(*shlib, "import"))
			results.push_back(imp);
		results.push_back(archiver.create_target(objs));
	}
	else if(type==MODULE)
		results.push_back(linker.create_target(objs, "shared"));
	else
		results.push_back(linker.create_target(objs));

	for(Target *r: results)
	{
		build_graph.add_primary_target(*r);
		if(install)
			build_graph.add_installed_target(*r);
	}
}

vector<FileTarget *> BinaryComponent::create_sources() const
{
	const Toolchain &toolchain = package.get_builder().get_toolchain();
	const Toolchain &pkg_tools = package.get_toolchain();

	vector<FileTarget *> targets;
	for(const FS::Path &s: collect_source_files())
	{
		string ext = FS::extpart(FS::basename(s));
		Target *src = 0;

		Tool *tool = pkg_tools.get_tool_for_suffix(ext, true);
		if(tool)
			src = tool->create_source(*this, s);
		if(!src)
		{
			tool = toolchain.get_tool_for_suffix(ext, true);
			if(tool)
				src = tool->create_source(*this, s);
		}

		if(FileTarget *file = dynamic_cast<FileTarget *>(src))
		{
			targets.push_back(file);
			if(type==LIBRARY)
				create_install(*file);
		}
	}

	return targets;
}

vector<Target *> BinaryComponent::extract_group(vector<FileTarget *> &targets, const Tool &tool)
{
	Tool::ProcessingUnit processing_unit = tool.get_processing_unit();
	vector<Target *> group;
	FS::Path group_dir;
	for(auto i=targets.begin(); i!=targets.end(); )
	{
		const FS::Path &path = (*i)->get_path();
		string ext = FS::extpart(FS::basename(path));
		FS::Path dir = FS::dirname(path);
		if(tool.accepts_suffix(ext) && (processing_unit!=Tool::DIRECTORY || group.empty() || dir==group_dir))
		{
			group.push_back(*i);
			group_dir = dir;
			i = targets.erase(i);
		}
		else
			++i;
	}

	return group;
}

void BinaryComponent::create_install(FileTarget &target) const
{
	BuildGraph &build_graph = package.get_builder().get_build_graph();
	auto add_func = (install ? &BuildGraph::add_installed_target : &BuildGraph::add_staged_target);

	if(target.is_installable())
		(build_graph.*add_func)(target);

	for(Target *s: target.get_side_effects())
		if(dynamic_cast<FileTarget &>(*s).is_installable())
			(build_graph.*add_func)(*s);
}


BinaryComponent::Loader::Loader(BinaryComponent &c):
	DataFile::DerivedObjectLoader<BinaryComponent, Component::Loader>(c)
{
	add("use", &Loader::use);
}

void BinaryComponent::Loader::use(const string &n)
{
	const BinaryComponent *comp = dynamic_cast<const BinaryComponent *>(&obj.package.get_component(n));
	if(!comp || comp->type!=LIBRARY)
		throw logic_error(n+" is not a library");

	obj.uses.push_back(comp);
}
