#include <deque>
#include <set>
#include <msp/core/algorithm.h>
#include <msp/core/except.h>
#include <msp/core/maputils.h>
#include <msp/datafile/parser.h>
#include <msp/fs/dir.h>
#include <msp/fs/stat.h>
#include <msp/fs/utils.h>
#include <msp/io/buffered.h>
#include <msp/io/file.h>
#include <msp/io/print.h>
#include <msp/strings/format.h>
#include <msp/strings/utils.h>
#include <msp/time/timedelta.h>
#include <msp/time/utils.h>
#include "binarypackage.h"
#include "builder.h"
#include "installedfile.h"
#include "package.h"
#include "plugin.h"
#include "sharedlibrary.h"
#include "task.h"
#include "tool.h"
#include "virtualtarget.h"

using namespace std;
using namespace Msp;

Builder::Builder():
	package_manager(*this),
	native_arch(*this, string()),
	vfs(*this),
	build_graph(*this),
	logger(&default_logger)
{
	set_architecture(string());
}

Builder::~Builder()
{
	if(current_arch!=&native_arch)
		delete current_arch;
}

void Builder::load_plugins()
{
	using CreateFunc = Plugin *(Builder &);

	FS::Path plugins_dir = FS::get_sys_lib_dir();
	if(!FS::exists(plugins_dir))
		return;

	logger->log("files", "Traversing %s", plugins_dir);
	vector<LoadedPlugin> unordered_plugins;
	for(const string &f: list_filtered(plugins_dir, "\\.dlm$"))
	{
		LoadedPlugin plugin;
		plugin.path = plugins_dir/f;

		try
		{
			plugin.module = new Module(plugin.path.str());
		}
		catch(const exception &exc)
		{
			logger->log("plugins", "Failed to load plugin %s: %s", f, exc.what());
			continue;
		}

		try
		{
			CreateFunc *create_func = reinterpret_cast<CreateFunc *>(plugin.module->get_symbol("create_plugin"));
			plugin.plugin = create_func(*this);
			if(plugin.plugin)
			{
				logger->log("plugins", "Loaded plugin %s", f);
				unordered_plugins.emplace_back(move(plugin));
				continue;
			}
			else
				logger->log("plugins", "Plugin %s refused to initialize", f);
		}
		catch(const exception &exc)
		{
			logger->log("plugins", "Failed to initialize plugin %s: %s", f, exc.what());
		}
	}

	add_plugins(unordered_plugins);
}

void Builder::add_plugins(vector<LoadedPlugin> &unordered_plugins)
{
	auto have_plugin = [this](const string &r){
		return any_of(plugins.begin(), plugins.end(), [&r](const LoadedPlugin &p){ return FS::basepart(FS::basename(p.path))==r; });
	};

	while(!unordered_plugins.empty())
	{
		bool any_added = false;
		for(auto i=unordered_plugins.begin(); i!=unordered_plugins.end(); )
		{
			const vector<string> &required = i->plugin->get_required_plugins();
			if(all_of(required.begin(), required.end(), have_plugin))
			{
				plugins.push_back(move(*i));
				i = unordered_plugins.erase(i);
				any_added = true;
			}
			else
				++i;
		}

		if(!any_added)
			break;
	}

	for(const LoadedPlugin &p: unordered_plugins)
	{
		vector<string> missing;
		for(const string &r: p.plugin->get_required_plugins())
			if(!have_plugin(r))
				missing.push_back(r);
		logger->log("plugins", "Missing required plugins for plugin %s: %s",
			FS::basename(p.path), join(missing.begin(), missing.end()));
	}
}

void Builder::set_architecture(const string &name)
{
	if(current_arch!=&native_arch)
		delete current_arch;

	if(name.empty())
		current_arch = &native_arch;
	else
		current_arch = new Architecture(*this, name);

	if(auto_prefix)
		update_auto_prefix();
}

vector<string> Builder::get_build_types() const
{
	vector<string> keys;
	keys.reserve(build_types.size());
	for(const auto &kvp: build_types)
		keys.push_back(kvp.first);
	return keys;
}

const BuildType &Builder::get_build_type() const
{
	if(!build_type)
		throw invalid_state("no build type");
	return *build_type;
}

void Builder::set_build_type(const string &name)
{
	build_type = &get_item(build_types, name);
}

void Builder::set_prefix(const FS::Path &p)
{
	auto_prefix = false;
	prefix = p;
}

void Builder::set_temp_directory(const FS::Path &p)
{
	tempdir = p;
}

void Builder::update_auto_prefix()
{
	if(current_arch->is_native())
		prefix = FS::get_home_dir()/"local";
	else
		prefix = FS::get_home_dir()/"local"/current_arch->get_name();
}

void Builder::add_default_tools()
{
	for(const LoadedPlugin &p: plugins)
		p.plugin->add_tools(toolchain, *current_arch);

	auto i = find_if(toolchain.get_toolchains(), [](const Toolchain *tc){ return (tc->has_tool("CC") || tc->has_tool("CXX")); });
	if(i!=toolchain.get_toolchains().end())
	{
		current_arch->refine((*i)->get_name());
		if(auto_prefix)
			update_auto_prefix();
	}
}

void Builder::set_logger(const Logger *l)
{
	logger = (l ? l : &default_logger);
}

vector<string> Builder::collect_problems() const
{
	vector<string> target_problems;
	vector<string> problems;
	vector<const Package *> broken_packages;
	vector<const Component *> broken_components;
	vector<const Tool *> broken_tools;

	for(const auto &kvp: build_graph.get_targets())
		if(kvp.second->is_broken())
		{
			const Package *package = kvp.second->get_package();
			if(package && package->is_broken())
				collect_broken_packages(*package, broken_packages);

			const Component *component = kvp.second->get_component();
			if(component && component->is_broken() && !any_equals(broken_components, component))
			{
				broken_components.push_back(component);
				collect_broken_packages(component->get_package(), broken_packages);
				for(const Package *r: component->get_required_packages())
					if(r->is_broken())
						collect_broken_packages(*r, broken_packages);
			}

			const Tool *tool = kvp.second->get_tool();
			if(tool && tool->is_broken() && !any_equals(broken_tools, tool))
			{
				broken_tools.push_back(tool);
				for(const string &p: tool->get_problems())
					problems.push_back(format("%s: %s", tool->get_tag(), p));
			}

			for(const string &p: kvp.second->get_problems())
				target_problems.push_back(format("%s: %s", kvp.second->get_name(), p));
		}

	for(const Package *p: broken_packages)
	{
		for(const string &b: p->get_problems())
			problems.push_back(format("%s: %s", p->get_name(), b));

		for(const Component *c: broken_components)
			if(&c->get_package()==p)
			{
				for(const string &b: c->get_problems())
					problems.push_back(format("%s/%s: %s", p->get_name(), c->get_name(), b));
			}
	}

	problems.insert(problems.end(), make_move_iterator(target_problems.begin()), make_move_iterator(target_problems.end()));

	return problems;
}

void Builder::collect_broken_packages(const Package &pkg, vector<const Package *> &broken) const
{
	if(any_equals(broken, &pkg))
		return;

	broken.push_back(&pkg);

	for(const Package *r: pkg.get_required_packages())
		if(r->is_broken())
			collect_broken_packages(*r, broken);
}

void Builder::load_build_file(const FS::Path &fn, const Config::InputOptions *opts, bool all)
{
	IO::BufferedFile in(fn.str());

	get_logger().log("files", "Reading %s", fn);

	DataFile::Parser parser(in, fn.str());
	Loader loader(*this, opts, all);
	loader.load(parser);
}

void Builder::save_caches()
{
	for(const auto &kvp: package_manager.get_packages())
		kvp.second->save_caches();
}

int Builder::build(unsigned jobs, bool dry_run, bool show_progress)
{
	unsigned total = build_graph.count_rebuild_targets();
	Time::TimeStamp start_time = Time::now();

	if(!total)
	{
		get_logger().log("summary", "Already up to date");
		return 0;
	}
	get_logger().log("summary", "Will build %d target%s", total, (total!=1 ? "s" : ""));

	vector<Task *> tasks;

	unsigned count = 0;
	Time::TimeDelta sum_time;

	bool fail = false;
	bool finish = false;
	bool starved = false;

	while(!finish)
	{
		if(tasks.size()<jobs && !fail && !starved)
		{
			Target *tgt = build_graph.get_buildable_target();
			if(tgt)
			{
				if(tgt->get_tool())
				{
					if(show_progress)
						IO::print("\033[K");
					get_logger().log("tasks", "%-4s  %s", tgt->get_tool()->get_tag(), tgt->get_name());
				}
				Task *task = tgt->build();
				if(task)
				{
					get_logger().log("commands", "%s", task->get_command());
					if(dry_run)
					{
						task->signal_finished.emit(true);
						delete task;
					}
					else
					{
						task->start();
						tasks.push_back(task);
					}
				}

				if(show_progress)
					IO::print("%d of %d target%s built\033[1G", count, total, (total!=1 ? "s" : ""));
			}
			else if(tasks.empty())
				finish = true;
			else
				starved = true;
		}
		else
			Time::sleep(10*Time::msec);

		for(unsigned i=0; i<tasks.size();)
		{
			Task::Status status;
			if(jobs==1 || (tasks.size()==1 && starved))
				status = tasks[i]->wait();
			else
				status = tasks[i]->check();

			if(status!=Task::RUNNING)
			{
				++count;

				const vector<const FileTarget *> &targets = tasks[i]->get_targets();
				if(!targets.empty())
				{
					sum_time += tasks[i]->get_duration();
					get_logger().log("timings", "%s built in %s", targets.front()->get_name(), tasks[i]->get_duration());
				}

				delete tasks[i];
				tasks.erase(tasks.begin()+i);
				if(status==Task::ERROR)
					fail = true;
				if(tasks.empty() && fail)
					finish = true;
				starved = false;
			}
			else
				++i;
		}
	}

	if(show_progress)
		IO::print("\033[K");
	if(fail)
		get_logger().log("summary", "Build failed");
	else if(show_progress)
		get_logger().log("summary", "Build complete");

	Time::TimeStamp end_time = Time::now();
	get_logger().log("timings", "Build took %s, with a total %s spent on tasks", end_time-start_time, sum_time);

	return fail;
}

int Builder::clean(bool all, bool dry_run)
{
	// Cleaning doesn't care about ordering, so a simpler method can be used

	set<Target *> clean_tgts;
	deque<Target *> queue;
	queue.push_back(&build_graph.get_goals());

	while(!queue.empty())
	{
		Target *tgt = queue.front();
		queue.pop_front();

		if(tgt->is_buildable() && (tgt->get_package()==&package_manager.get_main_package() || all))
			clean_tgts.insert(tgt);

		for(Target *t: tgt->get_dependencies())
			if(!clean_tgts.count(t))
				queue.push_back(t);
	}

	for(Target *t: clean_tgts)
	{
		get_logger().log("tasks", "RM    %s", t->get_name());
		if(!dry_run)
			t->clean();
	}

	return 0;
}


Builder::LoadedPlugin::LoadedPlugin(LoadedPlugin &&other):
	path(move(other.path)),
	module(other.module),
	plugin(other.plugin)
{
	other.module = 0;
	other.plugin = 0;
}

Builder::LoadedPlugin &Builder::LoadedPlugin::operator=(LoadedPlugin &&other)
{
	delete plugin;
	delete module;
	path = move(other.path);
	module = other.module;
	plugin = other.plugin;
	other.module = 0;
	other.plugin = 0;
	return *this;
}

Builder::LoadedPlugin::~LoadedPlugin()
{
	delete plugin;
	delete module;
}


Builder::Loader::Loader(Builder &b, const Config::InputOptions *o, bool a):
	DataFile::ObjectLoader<Builder>(b),
	options(o),
	conf_all(a)
{
	add("architecture", &Loader::architecture);
	add("binary_package", &Loader::binpkg);
	add("build_type", &Loader::build_type);
	add("package", &Loader::package);

	if(!obj.top_loader)
		obj.top_loader = this;
	else if(!options && obj.top_loader!=this && obj.top_loader->conf_all)
		options = obj.top_loader->options;
}

Builder::Loader::~Loader()
{
	if(obj.top_loader==this)
		obj.top_loader = 0;
}

void Builder::Loader::architecture(const string &n)
{
	if(obj.current_arch->match_name(n))
		load_sub(*obj.current_arch);
}

void Builder::Loader::binpkg(const string &n)
{
	BinaryPackage *pkg = new BinaryPackage(obj, n);
	load_sub(*pkg);
}

void Builder::Loader::build_type(const string &n)
{
	BuildType btype(n);
	load_sub(btype);
	auto i = obj.build_types.insert({ n, btype }).first;
	if(!obj.build_type)
		obj.build_type = &i->second;
}

void Builder::Loader::package(const string &n)
{
	SourcePackage *pkg = new SourcePackage(obj, n, get_source());

	load_sub(*pkg, options);

	if(obj.build_type)
		pkg->set_build_type(*obj.build_type);
}
