#ifndef BUILDER_H_
#define BUILDER_H_

#include <map>
#include <string>
#include <msp/core/module.h>
#include <msp/datafile/loader.h>
#include <msp/fs/path.h>
#include "architecture.h"
#include "buildgraph.h"
#include "buildtype.h"
#include "config.h"
#include "libbuilder_api.h"
#include "logger.h"
#include "packagemanager.h"
#include "sourcepackage.h"
#include "target.h"
#include "toolchain.h"
#include "virtualfilesystem.h"

class Package;
class Plugin;

/**
This class ties everything else together.  It also contains code for loading
build files and supervising the build process.
*/
class LIBBUILDER_API Builder
{
private:
	class Loader: public Msp::DataFile::ObjectLoader<Builder>
	{
	private:
		const Config::InputOptions *options;
		bool conf_all;

	public:
		Loader(Builder &, const Config::InputOptions * = 0, bool = false);
		~Loader();

	private:
		void architecture(const std::string &);
		void binpkg(const std::string &);
		void build_type(const std::string &);
		void package(const std::string &);
	};

private:
	struct LoadedPlugin
	{
		Msp::FS::Path path;
		Msp::Module *module = nullptr;
		Plugin *plugin = nullptr;

		LoadedPlugin() = default;
		LoadedPlugin(LoadedPlugin &&);
		LoadedPlugin &operator=(LoadedPlugin &&);
		~LoadedPlugin();
	};

	std::vector<LoadedPlugin> plugins;
	PackageManager package_manager;
	SourcePackage::ComponentRegistry component_registry;

	Architecture native_arch;
	Architecture *current_arch = 0;
	std::map<std::string, BuildType> build_types;
	BuildType *build_type = 0;
	Toolchain toolchain;
	VirtualFileSystem vfs;
	BuildGraph build_graph;
	Logger default_logger;
	const Logger *logger;

	bool auto_prefix = true;
	Msp::FS::Path prefix;
	Msp::FS::Path tempdir = "temp";

	Loader *top_loader = 0;

public:
	Builder();
	~Builder();

	void load_plugins();

	template<typename... T>
	void load_plugins();

private:
	void add_plugins(std::vector<LoadedPlugin> &);

public:
	PackageManager &get_package_manager() { return package_manager; }
	SourcePackage::ComponentRegistry &get_component_registry() { return component_registry; }

	template<typename F>
	void call_plugins(F) const;

	void set_architecture(const std::string &);
	const Architecture &get_current_arch() const { return *current_arch; }
	const Architecture &get_native_arch() const { return native_arch; }
	void set_build_type(const std::string &);
	std::vector<std::string> get_build_types() const;
	const BuildType &get_build_type() const;
	BuildGraph &get_build_graph() { return build_graph; }
	void set_prefix(const Msp::FS::Path &);
	void set_temp_directory(const Msp::FS::Path &);
	const Msp::FS::Path &get_prefix() const { return prefix; }
	const Msp::FS::Path &get_temp_directory() const { return tempdir; }

private:
	void update_auto_prefix();

public:
	void add_default_tools();
	const Toolchain &get_toolchain() const { return toolchain; }
	VirtualFileSystem &get_vfs() { return vfs; }
	void set_logger(const Logger *);
	const Logger &get_logger() const { return *logger; }

	std::vector<std::string> collect_problems() const;
private:
	void collect_broken_packages(const Package &, std::vector<const Package *> &) const;

public:
	/** Loads a build file.  If opts is not null, it is used to configure any
	packages loaded from this file.  If all is true, external packages are also
	configured. */
	void load_build_file(const Msp::FS::Path &, const Config::InputOptions *opts = 0, bool all = false);

	/** Saves package configuration and dependency caches. */
	void save_caches();

	/** Builds the goal targets.  The build graph must be prepared first. */
	int build(unsigned jobs = 1, bool dry_run = false, bool show_progress = false);

	/** Cleans buildable targets.  If all is true, cleans all packages.
	Otherwise cleans only the default package. */
	int clean(bool all = false, bool dry_run = false);

	int do_create_makefile();
};

template<typename... T>
void Builder::load_plugins()
{
	Plugin *raw_plugins[] = { new T(*this)... };
	std::vector<LoadedPlugin> pending_plugins;
	for(Plugin *p: raw_plugins)
	{
		LoadedPlugin plugin;
		plugin.plugin = p;
		pending_plugins.emplace_back(std::move(plugin));
	}
	add_plugins(pending_plugins);
}

template<typename F>
void Builder::call_plugins(F func) const
{
	for(const LoadedPlugin &p: plugins)
		func(*p.plugin);
}

#endif
