#ifndef BUILDGRAPH_H_
#define BUILDGRAPH_H_

#include <map>
#include <string>
#include "libbuilder_api.h"

class Builder;
class Target;

/**
Manages a graph of targets.
*/
class LIBBUILDER_API BuildGraph
{
private:
	Builder &builder;
	std::map<std::string, Target *> targets;
	Target *goals;

public:
	BuildGraph(Builder &);
	~BuildGraph();

	/** Looks up a target by name.  Returns 0 if no such target exists. */
	Target *get_target(const std::string &) const;

	const std::map<std::string, Target *> &get_targets() const { return targets; }

	/** Adds a target.  It can later be retrieved by name.  Called from Target
	constructor. */
	void add_target(Target *);

	/** Adds a target that is a primary build goal.  Such targets will be added
	as dependencies of the "world" virtual target.  If the target belongs to a
	default component of the main package, it's also added to the "default"
	virtual target. */
	void add_primary_target(Target &);

	/** Adds a target that will be installed.  A new InstalledFile target is
	created and added as a dependency to the "install" virtual target. */
	void add_installed_target(Target &);

	void add_staged_target(Target &);

	/** Adds a target as a toplevel goal.  These are stored as dependencies of
	the "goals" virtual target. */
	void add_goal(Target &);

	Target &get_goals() const { return *goals; }

	/** Prepares all toplevel goals for building.  If no goals are defined, the
	"default" target is added as a goal. */
	void prepare();

	/** Marks all buildable targets to be rebuilt.  The graph must be prepared
	first. */
	void force_full_rebuild();

	/** Returns the number of targets that are going to be rebuilt.  The graph
	must be prepared first. */
	unsigned count_rebuild_targets() const;

	/** Returns a target that can be built and is needed for building the goal
	targets.  Null */
	Target *get_buildable_target() const;
};

#endif
