#ifndef BUILDINFO_H_
#define BUILDINFO_H_

#include <string>
#include <vector>
#include <msp/datafile/objectloader.h>
#include <msp/fs/path.h>
#include "libbuilder_api.h"

/**
Stores information about compiler command line parameters in a more abstract
form.  Allows combining with other BuildInfos to support package dependencies.
*/
class LIBBUILDER_API BuildInfo
{
public:
	enum LibraryMode
	{
		FORCE_STATIC,  //< Only accept static libraries
		STATIC,        //< Prefer static libraries but accept dynamic as well
		DYNAMIC,       //< Prefer dynamic libraries but accept static as well
		FORCE_DYNAMIC  //< Only accept dynamic libraries
	};

	enum RuntimePathMode
	{
		NO_RPATH,  //< Do not record rpath in binaries
		RELATIVE,  //< Record relative rpath in binaries
		ABSOLUTE   //< Record absolute rpath in binaries
	};

	class LIBBUILDER_API Loader: public Msp::DataFile::ObjectLoader<BuildInfo>
	{
	public:
		Loader(BuildInfo &);
	private:
		void incpath(const std::string &);
		void define(const std::string &, const std::string &);
		void keep_symbol(const std::string &);
		void libmode_for_lib(const std::string &, LibraryMode);
		void libpath(const std::string &);
		void library(const std::string &);
		void local_incpath(const std::string &);
		void standard(Msp::DataFile::Symbol, const std::string &);
		void sysroot(const std::string &);
	};
	
	enum UpdateLevel
	{
		LOCAL,       //< Include all information
		DEPENDENCY,  //< Include all but code generation options
		CHAINED      //< Include only compilation options
	};

	struct LIBBUILDER_API LanguageStandard
	{
		std::string type;
		unsigned year = 0;

		LanguageStandard() = default;
		LanguageStandard(const std::string &);

		std::string str() const;
	};

	/**
	A wrapper which tracks the set status of the wrapped variable.  A default
	value may be provided in initialization without causing it to be treated as
	set.  Assigning from a raw value flags the Tracked object as set.  Assigning
	from another Tracked object will only change the value of the target if the
	source is set.  
	*/
	template<typename T>
	class Tracked
	{
	public:
		using LoadType = T;

	private:
		T value{};
		bool set = false;

	public:
		Tracked() = default;
		Tracked(T v): value(v) { }
		Tracked(const Tracked &t) = default;
		Tracked &operator=(const Tracked &v) { if(v.set) { value = v.value; set = true; } return *this; }

		Tracked &operator=(const T &v) { value = v; set = true; return *this; }
		operator const T &() const { return value; }
	};

	Tracked<Msp::FS::Path> sysroot;
	std::map<std::string, std::string> defines;
	std::vector<Msp::FS::Path> incpath;
	std::vector<Msp::FS::Path> local_incpath;
	std::vector<Msp::FS::Path> libpath;
	std::vector<std::string> libs;
	Tracked<LibraryMode> libmode = DYNAMIC;
	Tracked<RuntimePathMode> rpath_mode = NO_RPATH;
	std::map<std::string, LibraryMode> libmodes;
	std::vector<std::string> keep_symbols;
	std::map<std::string, LanguageStandard> standards;
	Tracked<bool> threads = false;
	Tracked<bool> debug = false;
	Tracked<int> optimize = 0;
	Tracked<bool> strip = false;
	Tracked<unsigned> warning_level = 0;
	Tracked<bool> fatal_warnings = false;

	/** Returns the library mode for linking a particular library.  If no mode
	has been specified for that library, the the global library mode is
	returned. */
	LibraryMode get_libmode_for(const std::string &) const;

	/** Updates the BuildInfo from another one.  Lists are concatenated, with
	the first occurrence of each item preserved.  Scalars are overwritten.
	
	The update level determines what information is updated. */
	void update_from(const BuildInfo &, UpdateLevel = LOCAL);
};

#endif
