#include <deque>
#include <msp/core/algorithm.h>
#include <msp/fs/dir.h>
#include <msp/fs/stat.h>
#include <msp/fs/utils.h>
#include <msp/strings/format.h>
#include <msp/strings/utils.h>
#include "builder.h"
#include "component.h"
#include "sourcepackage.h"

using namespace std;
using namespace Msp;

void Component::prepare()
{
	for(Package *r: required_pkgs)
	{
		r->prepare();
		broken |= r->is_broken();
	}

	if(!sources.empty())
	{
		for(auto i=sources.begin(); i!=sources.end(); ++i)
		{
			FS::Path s = *i;
			if(!FS::is_dir(s))
				s = FS::dirname(s);

			if(i==sources.begin())
				source_base_path = s;
			else
				source_base_path = FS::common_ancestor(source_base_path, s);
		}
	}

	if(!problems.empty())
		broken = true;
}

void Component::create_build_info()
{
	BuildInfo final_build_info;
	string build_macro = toupper(name)+"_BUILD";
	for(char &c: build_macro)
		if(!isalnum(static_cast<unsigned char>(c)))
			c = '_';
	final_build_info.defines[build_macro] = "1";

	const Package::Requirements &pkg_reqs = package.get_required_packages();
	Package::Requirements direct_reqs = required_pkgs;
	direct_reqs.insert(direct_reqs.end(), pkg_reqs.begin(), pkg_reqs.end());
	for(Package *r: direct_reqs)
		final_build_info.update_from(r->get_exported_build_info(), BuildInfo::DEPENDENCY);

	Package::Requirements all_reqs = direct_reqs;
	deque<Package *> queue(direct_reqs.begin(), direct_reqs.end());
	while(!queue.empty())
	{
		Package *req = queue.front();
		queue.pop_front();

		for(Package *r: req->get_required_packages())
			if(!any_equals(all_reqs, r))
			{
				final_build_info.update_from(r->get_exported_build_info(), BuildInfo::CHAINED);
				all_reqs.push_back(r);
				queue.push_back(r);
			}
	}

	final_build_info.update_from(package.get_build_info());
	final_build_info.update_from(build_info);
	build_info = final_build_info;

	for(FS::Path &p: build_info.incpath)
		p = (package.get_source_directory()/p).str();
	for(FS::Path &p: build_info.libpath)
		p = (package.get_source_directory()/p).str();
}

FS::Path Component::get_temp_directory() const
{
	return package.get_temp_directory()/name;
}

string Component::flatten_source_path(const FS::Path &source) const
{
	FS::Path temp_dir = get_temp_directory();
	FS::Path rel_src;
	if(FS::descendant_depth(source, temp_dir)>=0)
		rel_src = FS::relative(source, temp_dir);
	else
		rel_src = FS::relative(source, source_base_path);

	string fn;
	for(const string &c: rel_src)
		if(c!=".")
			append(fn, "_", c);

	return fn;
}

BuildInfo Component::get_build_info_for_path(const FS::Path &path) const
{
	// XXX Cache these and check that the directories actually exist before adding them
	BuildInfo binfo = build_info;

	if(!overlays.empty())
	{
		FS::Path dir = FS::dirname(path);
		string last = FS::basename(dir);
		if(any_equals(overlays, last))
			dir = FS::dirname(dir);

		if(any_equals(sources, dir))
		{
			binfo.local_incpath.push_back(dir);
			for(const string &o: overlays)
				binfo.local_incpath.push_back(dir/o);
		}
	}
	return binfo;
}

vector<FS::Path> Component::collect_source_files() const
{
	vector<FS::Path> files;
	for(const FS::Path &p: sources)
	{
		if(FS::is_dir(p))
		{
			vector<FS::Path> dirs;
			dirs.reserve(1+overlays.size());
			dirs.push_back(p);
			for(const string &o: overlays)
			{
				FS::Path opath = p/o;
				if(FS::is_dir(opath))
					dirs.push_back(opath);
			}
			set<string> overlay_files;
			for(auto j=dirs.begin(); j!=dirs.end(); ++j)
			{
				package.get_builder().get_logger().log("files", "Traversing %s", *j);
				for(const string &f: list_files(*j))
				{
					if(j!=dirs.begin())
					{
						if(overlay_files.count(f))
							continue;
						overlay_files.insert(f);
					}
					FS::Path fn = *j/f;
					if(!FS::is_dir(fn))
						files.push_back(fn);
				}
			}
		}
		else
		{
			files.push_back(p);
			for(const string &o: overlays)
			{
				FS::Path opath = FS::dirname(p)/o/FS::basename(p);
				if(FS::is_reg(opath))
					files.push_back(opath);
			}
		}
	}

	return files;
}


Component::Loader::Loader(Component &c):
	DataFile::ObjectLoader<Component>(c),
	ConditionalLoader(c.package, format("%s/%s", c.package.get_name(), c.name))
{
	add("overlay",         &Loader::overlay);
	add("source",          &Loader::source);
	add("install",         &Component::install);
	add("install_map",     &Loader::install_map);
	add("build_info",      &Loader::build_info);
	add("require",         &Loader::require);
	add("default",         &Component::deflt);
}

void Component::Loader::build_info()
{
	load_sub(obj.build_info);
}

void Component::Loader::install_map()
{
	load_sub(obj.install_map, obj.package.get_source_directory());
}

void Component::Loader::overlay(const string &o)
{
	obj.overlays.push_back(o);
}

void Component::Loader::require(const string &n)
{
	Package *req = obj.package.get_builder().get_package_manager().find_package(n);
	if(req)
		obj.required_pkgs.push_back(req);
	else
		obj.problems.push_back(format("Required package %s not found", n));
}

void Component::Loader::source(const string &s)
{
	FS::Path src_path = obj.package.get_source_directory()/s;
	if(!FS::exists(src_path))
		throw IO::file_not_found(src_path.str());
	obj.sources.push_back(src_path);
}
