#include <msp/core/algorithm.h>
#include <msp/fs/stat.h>
#include <msp/fs/utils.h>
#include <msp/strings/format.h>
#include <msp/strings/utils.h>
#include <msp/time/utils.h>
#include "builder.h"
#include "filetarget.h"
#include "sourcepackage.h"
#include "task.h"
#include "tool.h"

using namespace std;
using namespace Msp;

FileTarget::FileTarget(Builder &b, const FS::Path &a):
	FileTarget(b, 0, a)
{ }

FileTarget::FileTarget(Builder &b, const SourcePackage &p, const FS::Path &a):
	FileTarget(b, &p, a)
{ }

FileTarget::FileTarget(Builder &b, const SourcePackage *p, const FS::Path &a):
	Target(b, generate_name(b, p, a)),
	path(a)
{
	package = p;

	builder.get_vfs().register_path(path, this);

	stat();
}

void FileTarget::stat()
{
	if(FS::Stat st = FS::lstat(path))
	{
		mtime = st.get_modify_time();
		size = st.get_size();
	}
}

string FileTarget::generate_name(Builder &builder, const SourcePackage *pkg, const FS::Path &path)
{
	if(pkg && FS::descendant_depth(path, pkg->get_source_directory())>=0)
	{
		FS::Path relpath = FS::relative(path, pkg->get_source_directory());
		return format("<%s>%s", pkg->get_name(), relpath.str().substr(1));
	}
	else if(FS::descendant_depth(path, builder.get_prefix())>=0)
	{
		FS::Path relpath = FS::relative(path, builder.get_prefix());
		return "<prefix>"+relpath.str().substr(1);
	}

	return path.str();
}

void FileTarget::touch()
{
	mtime = Time::now();
	modified();
	signal_bubble_rebuild.emit();
}

Target *FileTarget::resolve_transitive_dependency(Target &from, Target &dep) const
{
	FileTarget *from_file = dynamic_cast<FileTarget *>(&from);
	FileTarget *dep_file = dynamic_cast<FileTarget *>(&dep);
	if(!from_file || !dep_file)
		return &dep;

	const SourcePackage *from_pkg = from.get_package();
	if(dep_file->get_package()!=from_pkg || FS::descendant_depth(dep_file->get_path(), from_pkg->get_source_directory())<0)
		return &dep;

	/* Adjust the path of the dependency to account for the install location
	of the dependent target. */
	const Component *from_comp = from_file->get_real_target()->get_component();
	const Component *dep_comp = dep_file->get_component();
	FS::Path from_inst = from_comp->get_install_map().get_install_location(*from_file->get_real_target());
	FS::Path dep_inst = dep_comp->get_install_map().get_install_location(*dep_file);
	FS::Path displaced = FS::dirname(from_file->get_path())/FS::relative(dep_inst, from_inst)/FS::basename(dep_file->get_path());
	if(Target *result = builder.get_vfs().get_target(displaced))
		return result;

	/* If the target was in an overlay directory and the displaced
	dependency is not found, try removing the overlay from the path. */
	string last_dir = FS::basename(FS::dirname(displaced));
	if(any_equals(dep_comp->get_overlays(), last_dir))
	{
		displaced = displaced.subpath(0, displaced.size()-2)/FS::basename(dep_file->get_path());
		return builder.get_vfs().get_target(displaced);
	}

	return nullptr;
}

void FileTarget::check_rebuild()
{
	if(!tool || needs_rebuild())
		return;

	if(!mtime)
		mark_rebuild("Does not exist");
	else
	{
		for(Target *d: depends)
		{
			FileTarget *ft = dynamic_cast<FileTarget *>(d);
			if(ft && ft->get_mtime()>mtime)
				mark_rebuild(d->get_name()+" has changed");
			else if(d->needs_rebuild())
				mark_rebuild(d->get_name()+" needs rebuilding");
			if(needs_rebuild())
				break;
		}
	}

	if(!needs_rebuild())
	{
		// Some side effects might not exist
		auto i = find_if(side_effects, [](const Target *s){ return s->needs_rebuild(); });
		if(i!=side_effects.end())
			mark_rebuild((*i)->get_name()+" needs rebuilding");
		if(primary_target && primary_target->needs_rebuild())
			mark_rebuild(primary_target->get_name()+" needs rebuilding");
	}

	if(!needs_rebuild() && package)
	{
		if(package->get_config().get_mtime()>mtime)
			mark_rebuild("Package options changed");

		if(tool->get_executable())
		{
			string build_sig = create_build_signature();
			if(package->get_cache().has_key(this, "build_sig"))
			{
				if(package->get_cache().get_value(this, "build_sig")!=build_sig)
					mark_rebuild("Build signature changed");
			}
		}
	}
}

string FileTarget::create_build_signature() const
{
	if(!package)
		return string();

	const BuildInfo &binfo = (component ? component->get_build_info() : package->get_build_info());
	vector<string> sigs;

	if(arch_in_build_sig)
		if(const Architecture *arch = tool->get_architecture())
			sigs.push_back(arch->get_name());

	sigs.push_back(tool->create_build_signature(binfo));

	if(nested_build_sig && component)
	{
		vector<const Tool *> seen_tools;
		vector<string> tool_sigs;
		for(Target *d: depends)
			if(const Tool *t = d->get_tool())
				if(d->get_component()==component && !any_equals(seen_tools, t))
				{
					seen_tools.push_back(t);
					tool_sigs.push_back(t->create_build_signature(binfo));
				}

		sort(tool_sigs);
		sigs.insert(sigs.end(), make_move_iterator(tool_sigs.begin()), make_move_iterator(tool_sigs.end()));
	}

	return join(sigs.begin(), sigs.end(), ";");
}

void FileTarget::build(Task &task)
{
	task.add_target(*this);
	task.set_unlink(true);
}

void FileTarget::build_finished(bool success)
{
	if(success)
	{
		stat();
		if(package)
		{
			string build_sig = create_build_signature();
			if(!build_sig.empty())
				package->get_cache().set_value(this, "build_sig", build_sig);
		}
	}

	Target::build_finished(success);
}

void FileTarget::clean()
{
	if(mtime)
	{
		FS::unlink(path);
		mtime = Time::TimeStamp();
		size = 0;
		check_rebuild();
	}
}
