#include <msp/fs/utils.h>
#include <msp/strings/format.h>
#include <msp/strings/utils.h>
#include "builder.h"
#include "executable.h"
#include "sourcegenerator.h"
#include "sourcepackage.h"
#include "templatefile.h"

using namespace std;
using namespace Msp;

SourceGenerator::SourceGenerator(Builder &b, const SourcePackage &p, const string &t):
	Tool(b, t),
	package(p)
{
	set_run_external(&_run);
}

Target *SourceGenerator::create_source(const Component &comp, const FS::Path &path) const
{
	return new TemplateFile(builder, comp, path);
}

Target *SourceGenerator::create_target(const vector<Target *> &sources, const string &)
{
	if(sources.empty())
		throw invalid_argument("SourceGenerator::create_target");
	if(out_suffixes.empty())
		throw logic_error("No output suffixes");

	TemplateFile &tmpl = dynamic_cast<TemplateFile &>(*sources.front());
	const Component *comp = tmpl.get_component();
	string subdir = "generated";
	string base;
	if(processing_unit==COMPONENT)
		base = comp->get_name();
	else
	{
		string flat_dir = comp->flatten_source_path(FS::dirname(tmpl.get_path()));
		if(processing_unit==ONE_FILE)
		{
			append(subdir, "_", flat_dir);
			base = FS::basepart(FS::basename(tmpl.get_path()));
		}
		else if(processing_unit==DIRECTORY)
			base = flat_dir;
	}

	Target *primary = 0;
	for(const string &s: out_suffixes)
	{
		Tool *tool = builder.get_toolchain().get_tool_for_suffix(s, true);
		if(tool)
		{
			FS::Path fn = comp->get_temp_directory()/subdir/(base+s);
			Target *target = tool->create_source(*comp, fn);
			target->set_tool(*this);
			for(Target *t: sources)
				target->add_dependency(*t);
			if(primary)
				primary->add_side_effect(*target);
			else
				primary = target;
		}
		else
			throw runtime_error("No tool found for suffix "+s);
	}

	return primary;
}

ExternalTask::Arguments SourceGenerator::_run(const SourceFile &out_src, FS::Path &work_dir)
{
	const SourceGenerator &tool = dynamic_cast<const SourceGenerator &>(*out_src.get_tool());

	vector<string> args;
	args.push_back(tool.get_executable()->get_path().str());
	args.insert(args.end(), tool.arguments.begin(), tool.arguments.end());

	for(const Target *d: out_src.get_dependencies())
		if(const TemplateFile *tmpl = dynamic_cast<const TemplateFile *>(d))
			args.push_back(FS::relative(tmpl->get_path(), work_dir).str());

	if(!tool.out_argument.empty())
		args.push_back(tool.out_argument);
	args.push_back(FS::relative(out_src.get_path(), work_dir).str());

	return args;
}


SourceGenerator::Loader::Loader(SourceGenerator &sg):
	DataFile::ObjectLoader<SourceGenerator>(sg),
	ConditionalLoader(sg.package, format("%s/%s", sg.package.get_name(), sg.tag))
{
	add("argument",   &Loader::argument);
	add("arguments",  &Loader::arguments);
	add("command",    &Loader::command);
	add("in_suffix",  &Loader::in_suffix);
	add("out_argument", &SourceGenerator::out_argument);
	add("out_suffix", &Loader::out_suffix);
	add("processing_unit", static_cast<ProcessingUnit SourceGenerator::*>(&SourceGenerator::processing_unit));
}

void SourceGenerator::Loader::argument(const string &a)
{
	obj.arguments.push_back(a);
}

void SourceGenerator::Loader::arguments(const vector<string> &a)
{
	obj.arguments.insert(obj.arguments.end(), a.begin(), a.end());
}

void SourceGenerator::Loader::command(const string &c)
{
	if(c.find('/')!=string::npos)
		obj.set_command((obj.package.get_source_directory()/c).str());
	else
		obj.set_command(c);
}

void SourceGenerator::Loader::in_suffix(const string &s)
{
	obj.input_suffixes.push_back(s);
}

void SourceGenerator::Loader::out_suffix(const string &s)
{
	obj.out_suffixes.push_back(s);
}
