#include <cstdlib>
#include <msp/core/algorithm.h>
#include <msp/core/maputils.h>
#include <msp/fs/utils.h>
#include <msp/io/print.h>
#include <msp/strings/lexicalcast.h>
#include <msp/strings/utils.h>
#include "binarycomponent.h"
#include "binarypackage.h"
#include "builder.h"
#include "file.h"
#include "installcomponent.h"
#include "plugin.h"
#include "sourcearchivecomponent.h"
#include "sourcegenerator.h"
#include "sourcepackage.h"
#include "tool.h"

using namespace std;
using namespace Msp;

SourcePackage::SourcePackage(Builder &b, const string &n, const FS::Path &f):
	Package(b, n),
	source_dir(FS::dirname(f)),
	config(*this),
	cache(*this)
{
	config.load();

	build_file = builder.get_vfs().get_target(f);
	if(!build_file)
		build_file = new File(builder, *this, f);
	source_archive = new SourceArchiveComponent(*this);
	components.push_back(source_archive);
}

SourcePackage::~SourcePackage()
{
	for(Component *c: components)
		delete c;
}

FS::Path SourcePackage::get_temp_directory() const
{
	string subdir = builder.get_current_arch().get_name();
	if(build_type)
	{
		subdir += '.';
		subdir += build_type->get_name();
	}

	const FS::Path &temp = builder.get_temp_directory();
	if(temp.is_absolute())
		return temp/name/subdir;
	else
		return source_dir/temp/subdir;
}

FS::Path SourcePackage::get_output_directory() const
{
	const Architecture &arch = builder.get_current_arch();
	if(arch.is_native())
		return source_dir;
	else
		return source_dir/arch.get_name();
}

FS::Path SourcePackage::get_staging_directory() const
{
	return get_temp_directory()/"staging";
}

const Component &SourcePackage::get_component(const string &n) const
{
	auto i = find_if(components, [&n](const Component *c){ return c->get_name()==n; });
	if(i!=components.end())
		return **i;
	throw key_error(n);
}

bool SourcePackage::match_feature(const string &feat, const string *comp) const
{
	string value = config.get_option("with_"+feat).value;
	if(comp)
		return value==*comp;
	else
		return lexical_cast<bool>(value);
}

void SourcePackage::set_build_type(const BuildType &t)
{
	build_type = &t;
}

void SourcePackage::do_prepare()
{
	BuildInfo final_build_info;

	if(build_type)
		final_build_info.update_from(build_type->get_build_info());

	final_build_info.update_from(build_info);
	build_info = final_build_info;

	build_info.incpath.push_back(get_staging_directory()/"include");
	build_info.incpath.push_back(builder.get_prefix()/"include");
	build_info.libpath.push_back(builder.get_prefix()/"lib");

	for(const Feature &f: features)
	{
		string ident = "WITH_"+toupper(f.name);
		string value = config.get_option("with_"+f.name).value;

		if(f.choices.empty())
		{
			if(!lexical_cast<bool>(value))
				continue;
			value = "1";
		}

		build_info.defines[ident] = value;
		if(f.exported)
			export_binfo.defines[ident] = value;
	}

	for(Component *c: components)
	{
		c->prepare();
		c->create_build_info();

		c->update_exported_build_info(export_binfo);
	}

	cache.load();

	for(Component *c: components)
		c->create_targets();

	if(!export_binfo.libs.empty())
	{
		export_binfo.incpath.push_back((builder.get_prefix()/"include").str());
		export_binfo.libpath.push_back((builder.get_prefix()/"lib").str());
	}

	export_binfo.standards = build_info.standards;

	builder.call_plugins([this](const Plugin &p){ p.create_targets(*this); });
}

void SourcePackage::save_caches()
{
	config.save();
	cache.save();
}


SourcePackage::Loader::Loader(SourcePackage &p, const Config::InputOptions *o):
	DataFile::DerivedObjectLoader<SourcePackage, Package::Loader>(p),
	FeatureConditional(p, p.name),
	options(o)
{
	add("build_info",  &Loader::build_info);
	add("description", &SourcePackage::description);
	add("export",      &Loader::exported);
	add("feature",     &Loader::feature);
	add("generate",    &Loader::generate);
	add("install",     &Loader::component<InstallComponent>);
	add("interface_version", &Loader::interface_version);
	add("library",     &Loader::component<BinaryComponent, BinaryComponent::Type>, BinaryComponent::LIBRARY);
	add("module",      &Loader::component<BinaryComponent, BinaryComponent::Type>, BinaryComponent::MODULE);
	add("program",     &Loader::component<BinaryComponent, BinaryComponent::Type>, BinaryComponent::PROGRAM);
	add("source_archive", &Loader::source_archive);
	add("source_tarball", &Loader::source_archive);
	add("tarball",     &Loader::tarball);
	add("version",     &Loader::version);

	p.builder.get_component_registry().invoke_all(*this);
}

void SourcePackage::Loader::finish()
{
	/* Make sure the source tarball is last in the list so targets from all
	other components wil be created first */
	auto i = find(obj.components, obj.source_archive);
	if(i!=obj.components.end())
	{
		obj.components.erase(i);
		obj.components.push_back(obj.source_archive);
	}
}

void SourcePackage::Loader::exported()
{
	ExportLoader ldr(obj);
	load_sub_with(ldr);
}

void SourcePackage::Loader::feature(const string &n, const string &d)
{
	Feature feat(n);
	feat.description = d;
	load_sub(feat);
	obj.features.push_back(feat);

	const Config::Option &opt = obj.config.add_option(feat);
	if(options)
	{
		auto i = options->find(opt.name);
		if(i!=options->end())
			obj.config.set_option(opt.name, i->second);
	}
}

void SourcePackage::Loader::build_info()
{
	load_sub(obj.build_info);
}

void SourcePackage::Loader::generate(const string &tag)
{
	SourceGenerator *gen = new SourceGenerator(obj.builder, obj, tag);
	load_sub(*gen);
	obj.local_tools.add_tool(gen);
}

void SourcePackage::Loader::interface_version(const string &v)
{
	obj.interface_version = v;
	if(obj.version.empty())
		obj.version = v;
}

void SourcePackage::Loader::source_archive()
{
	load_sub(*obj.source_archive);
}

void SourcePackage::Loader::tarball(const string &)
{
	IO::print("%s: Deprecated tarball component ignored\n", get_source());
}

void SourcePackage::Loader::version(const string &v)
{
	obj.version = v;

	string::size_type i = 0;
	for(unsigned dots=0; i<obj.version.size(); ++i)
		if(obj.version[i]=='.' && ++dots>=2)
			break;
	obj.interface_version = obj.version.substr(0, i);
}


SourcePackage::ExportLoader::ExportLoader(SourcePackage &p):
	ObjectLoader<SourcePackage>(p)
{
	add("build_info", &ExportLoader::build_info);
}

void SourcePackage::ExportLoader::build_info()
{
	BuildInfo bi;
	load_sub(bi);
	obj.build_info.update_from(bi);
	obj.export_binfo.update_from(bi);
}
