#ifndef SOURCEPACKAGE_H_
#define SOURCEPACKAGE_H_

#include <algorithm>
#include <stdexcept>
#include <string>
#include <msp/core/typeregistry.h>
#include "buildinfo.h"
#include "cache.h"
#include "component.h"
#include "conditionalloader.h"
#include "config.h"
#include "feature.h"
#include "libbuilder_api.h"
#include "package.h"
#include "toolchain.h"

class BuildType;
class FileTarget;
class SourceArchiveComponent;

/**
A package that can be built by Builder.
*/
class LIBBUILDER_API SourcePackage: public Package
{
public:
	class Loader: public Msp::DataFile::DerivedObjectLoader<SourcePackage, Package::Loader>, public FeatureConditional
	{
		friend class SourcePackage;

	private:
		template<typename T>
		struct AddComponent
		{
			void operator()(const std::string &n, Loader &l) const { l.add(n, &Loader::component<T>); }
		};

		const Config::InputOptions *options;

	public:
		Loader(SourcePackage &, const Config::InputOptions *);
	private:
		void finish() override;

		void exported();
		void feature(const std::string &, const std::string &);
		template<typename C, typename... Args>
		void component(Args..., const std::string &);
		void build_info();
		void generate(const std::string &);
		void interface_version(const std::string &);
		void source_archive();
		void tarball(const std::string &);
		void version(const std::string &);
	};

	using ComponentRegistry = Msp::TypeRegistry<Loader::AddComponent, Loader &>;

private:
	class ExportLoader: public Msp::DataFile::ObjectLoader<SourcePackage>
	{
	public:
		ExportLoader(SourcePackage &);

	private:
		void build_info();
	};

	std::string version;
	std::string interface_version;
	std::string description;

	FileTarget *build_file;
	Msp::FS::Path source_dir;
	const BuildType *build_type = 0;
	Toolchain local_tools;
	std::vector<Feature> features;
	BuildInfo build_info;
	std::vector<Component *> components;
	SourceArchiveComponent *source_archive;
	Config config;
	mutable Cache cache;

public:
	SourcePackage(Builder &, const std::string &, const Msp::FS::Path &);
	~SourcePackage();

	const std::string &get_version() const { return version; }
	const std::string &get_interface_version() const { return interface_version; }
	const std::string &get_description() const { return description; }

	FileTarget &get_build_file() const { return *build_file; }
	const Msp::FS::Path &get_source_directory() const { return source_dir; }
	Msp::FS::Path get_temp_directory() const;
	Msp::FS::Path get_output_directory() const;
	Msp::FS::Path get_staging_directory() const;

	const Toolchain &get_toolchain() const { return local_tools; }
	const Component &get_component(const std::string &) const;
	const Config &get_config() const { return config; }
	bool match_feature(const std::string &, const std::string *) const;
	void set_build_type(const BuildType &);
	const BuildInfo &get_build_info() const { return build_info; }
private:
	void do_prepare() override;

public:
	Cache &get_cache() const { return cache; }
private:
	void save_caches() override;
};


template<typename C, typename... Args>
void SourcePackage::Loader::component(Args... args, const std::string &n)
{
	if(std::any_of(obj.components.begin(), obj.components.end(), [&n](const Component *c){ return c->get_name()==n; }))
		throw Msp::key_error(n);
	C *comp = new C(obj, n, args...);
	load_sub(*comp);
	obj.components.push_back(comp);
}

#endif
