#include <msp/core/algorithm.h>
#include <msp/fs/stat.h>
#include <msp/fs/utils.h>
#include <msp/strings/utils.h>
#include "builder.h"
#include "filetarget.h"
#include "sourcepackage.h"
#include "target.h"
#include "task.h"
#include "tool.h"

using namespace std;
using namespace Msp;

vector<Target *> Target::prepare_stack;

Target::Target(Builder &b, const string &n):
	builder(b),
	name(n)
{
	builder.get_build_graph().add_target(this);
}

void Target::add_dependency(Target &dep)
{
	if(&dep==this)
		throw invalid_argument("Target::add_depend");
	depends.push_back(&dep);
	if(state>PREPARING)
		dep.signal_bubble_rebuild.connect(sigc::mem_fun(this, &Target::check_rebuild));
}

void Target::add_transitive_dependency(Target &dep)
{
	if(&dep==this)
		throw invalid_argument("Target::add_transitive_dependency");
	trans_depends.push_back(&dep);
}

void Target::add_side_effect(Target &se)
{
	side_effects.push_back(&se);
	if(tool)
		se.set_tool(*tool);
	se.primary_target = this;
	/* Side effects are checked for rebuild after the primary target.  Recheck
	the primary if a side effect is marked for rebuild. */
	se.signal_bubble_rebuild.connect(sigc::mem_fun(this, &Target::check_rebuild));
}

bool Target::find_transitive_dependencies()
{
	vector<Target *> found_deps;
	for(Target *d: depends)
		if(!d->is_buildable())
			find_transitive_dependencies(*d, found_deps);

	bool any_added = false;
	for(Target *d: found_deps)
		if(!any_equals(depends, d))
		{
			any_added = true;
			add_dependency(*d);
			if(d->get_real_target()->is_buildable())
				d->signal_modified.connect([this]{ rescan_trans_deps = true; });
		}

	return any_added;
}

void Target::find_transitive_dependencies(Target &tgt, vector<Target *> &found_deps) const
{
	tgt.prepare();

	Target *rtgt = tgt.get_real_target();
	Dependencies deps_to_add = rtgt->get_transitive_dependencies();
	if(rtgt!=&tgt)
	{
		for(Target *&d: deps_to_add)
			d = resolve_transitive_dependency(tgt, *d);
	}

	for(Target *d: deps_to_add)
		if(d)
		{
			auto i = lower_bound(found_deps, d);
			if(i==found_deps.end() || *i!=d)
			{
				found_deps.insert(i, d);
				find_transitive_dependencies(*d, found_deps);
			}
		}
}

Target *Target::get_buildable_target()
{
	if(primary_target)
		return primary_target->get_buildable_target();
	if(!needs_rebuild())
		return 0;

	bool self_ok = state!=BUILDING;
	for(Target *d: depends)
	{
		// Avoid infinite recursion if a target depends on its own side effect
		if(any_equals(side_effects, d))
			continue;

		Target *tgt = d->get_buildable_target();
		if(tgt)
			return tgt;
		else if(d->needs_rebuild())
			self_ok = false;
	}

	if(self_ok)
	{
		if(rescan_trans_deps)
		{
			rescan_trans_deps = false;
			if(find_transitive_dependencies())
				return get_buildable_target();
		}

		return this;
	}

	return 0;
}

void Target::set_tool(Tool &t)
{
	tool = &t;
	for(Target *s: side_effects)
		s->set_tool(t);
}

void Target::collect_build_info(BuildInfo &binfo) const
{
	if(tool)
		binfo.update_from(tool->get_build_info());
	if(component)
		binfo.update_from(component->get_build_info());
	else if(package)
		binfo.update_from(package->get_build_info());
}

void Target::force_rebuild()
{
	if(!is_buildable())
		throw logic_error("Target::force_rebuild");
	mark_rebuild("Forced rebuild");
}

void Target::mark_rebuild(const string &reason)
{
	if(reason.empty())
		throw invalid_argument("No reason given for rebuilding "+name);

	state = REBUILD;
	rebuild_reason = reason;

	builder.get_logger().log("rebuild", "Rebuilding %s: %s", name, reason);

	signal_bubble_rebuild.emit();
}

void Target::prepare()
{
	if(primary_target && primary_target->state<PREPARING)
		primary_target->prepare();

	if(state>PREPARING)
		return;
	if(state==PREPARING)
	{
		auto i = find(prepare_stack, this);
		if(i!=prepare_stack.end())
		{
			string cycle;
			for(; i!=prepare_stack.end(); ++i)
				append(cycle, " -> ", (*i)->name);
			append(cycle, " -> ", name);
			builder.get_logger().log("problems", "Dependency cycle detected: %s", cycle);
			problems.push_back(format("Dependency cycle detected: %s", cycle));
		}
		else
		{
			builder.get_logger().log("problems", "Dependency cycle detected at %s", name);
			problems.push_back("Dependency cycle detected");
		}
		state = BROKEN;
		return;
	}

	PushPrepare _push(this);
	state = PREPARING;
	/* Prepare existing dependencies early, because their information may be
	needed to find other dependencies. */
	for(Target *d: depends)
		d->prepare();
	if(tool)
		tool->prepare();

	find_dependencies();
	find_transitive_dependencies();
	bool broken = !problems.empty();

	if(tool)
	{
		if(FileTarget *tool_exe = tool->get_executable())
			add_dependency(*tool_exe);
		broken |= tool->is_broken();

		// Only check package and component brokenness for buildable targets
		broken |= (package && package->is_broken());
		broken |= (component && component->is_broken());
	}

	/* Now that all dependencies are known, prepare them again.  This will do
	nothing to already prepared targets. */
	for(Target *d: depends)
	{
		d->prepare();
		broken |= d->is_broken();
	}
	for(Target *d: trans_depends)
		d->prepare();

	check_rebuild();
	if(broken)
		state = BROKEN;
	else if(state==PREPARING)
		state = UPTODATE;

	for(Target *d: depends)
		d->signal_bubble_rebuild.connect(sigc::mem_fun(this, &Target::check_rebuild));
}

Task *Target::build()
{
	if(primary_target)
		return primary_target->build();

	Task *task = tool->run(*this);
	task->signal_finished.connect(sigc::mem_fun(this, &Target::build_finished));
	state = BUILDING;

	build(*task);
	for(Target *s: side_effects)
		s->build(*task);

	return task;
}

void Target::build_finished(bool success)
{
	state = UPTODATE;
	if(success)
	{
		modified();
		for(Target *s: side_effects)
			s->build_finished(success);
		signal_modified.emit();
	}
}
