#include <msp/core/environ.h>
#include <msp/fs/stat.h>
#include <msp/fs/utils.h>
#include <msp/io/print.h>
#include <msp/strings/utils.h>
#include "builder.h"
#include "executable.h"
#include "importlibrary.h"
#include "sharedlibrary.h"
#include "staticlibrary.h"
#include "tool.h"
#include "virtualfilesystem.h"

using namespace std;
using namespace Msp;

FileTarget *VirtualFileSystem::get_target(const FS::Path &p) const
{
	auto i = targets.find(p.str());
	if(i!=targets.end())
		return static_cast<FileTarget *>(i->second);
	return 0;
}

void VirtualFileSystem::register_path(const FS::Path &path, FileTarget *t)
{
	targets.insert({ path, t });
	nonexistent.erase(path);
	builder.get_logger().log("vfs", "Path %s registered to %s", path, t->get_name());
}

FileTarget *VirtualFileSystem::find_header(const string &name, Tool *tool, const SearchPath &path, bool use_syspath)
{
	if(!tool)
		tool = builder.get_toolchain().get_tool_for_suffix(FS::extpart(FS::basename(name)), true);
	if(!tool)
		return 0;

	tool->prepare();

	SearchPath combined_path = path;
	if(use_syspath)
	{
		const SearchPath &syspath = tool->get_system_path();
		combined_path.insert(combined_path.end(), syspath.begin(), syspath.end());
	}

	for(const FS::Path &p: combined_path)
	{
		FS::Path filename = p/name;
		if(FileTarget *tgt = get_target(filename))
		{
			builder.get_logger().log("vfs", "Header %s found in %s as existing %s", name, p.str(), tgt->get_type());
			return tgt;
		}
		else if(file_exists(filename))
		{
			builder.get_logger().log("vfs", "Header %s found in %s", name, p.str());
			return dynamic_cast<FileTarget *>(tool->create_source(filename));
		}

		builder.get_logger().log("vfs", "Header %s not found in %s", name, p.str());
	}

	return 0;
}

FileTarget *VirtualFileSystem::find_library(const string &lib, const SearchPath &path, BuildInfo::LibraryMode mode, bool use_syspath)
{
	SearchPath combined_path = path;
	if(use_syspath)
	{
		Tool &linker = builder.get_toolchain().get_tool("LINK");
		linker.prepare();
		const SearchPath &syspath = linker.get_system_path();
		combined_path.insert(combined_path.end(), syspath.begin(), syspath.end());
	}

	const Architecture &arch = builder.get_current_arch();

	vector<string> shared_names;
	bool use_import_lib = false;
	if(mode!=BuildInfo::FORCE_STATIC)
	{
		shared_names = Pattern::apply_list(arch.get_patterns<ImportLibrary>(), lib);
		if(!(use_import_lib = !shared_names.empty()))
			shared_names = Pattern::apply_list(arch.get_patterns<SharedLibrary>(), lib);
	}

	vector<string> static_names;
	if(mode!=BuildInfo::FORCE_DYNAMIC)
		static_names = Pattern::apply_list(arch.get_patterns<StaticLibrary>(), lib);

	for(const FS::Path &p: combined_path)
	{
		const vector<string> *cur_names = (mode>=BuildInfo::DYNAMIC ? &shared_names : &static_names);
		for(auto j=cur_names->begin(); j!=cur_names->end(); )
		{
			FS::Path filename = p / *j;
			if(FileTarget *tgt = get_target(filename))
			{
				builder.get_logger().log("vfs", "Library %s (%s) found in %s as existing %s", lib, *j, p.str(), tgt->get_type());
				return tgt;
			}
			else if(file_exists(filename))
			{
				builder.get_logger().log("vfs", "Library %s (%s) found in %s", lib, *j, p.str());
				if(cur_names==&shared_names)
				{
					if(use_import_lib)
						return new ImportLibrary(builder, filename);
					return new SharedLibrary(builder, filename);
				}
				else
					return new StaticLibrary(builder, filename);
			}

			if(++j==cur_names->end())
			{
				if(mode==BuildInfo::DYNAMIC && cur_names==&shared_names)
					cur_names = &static_names;
				else if(mode==BuildInfo::STATIC && cur_names==&static_names)
					cur_names = &shared_names;
				else
					break;
				j = cur_names->begin();
			}
		}

		builder.get_logger().log("vfs", "Library %s not found in %s", lib, p.str());
	}

	return 0;
}

FileTarget *VirtualFileSystem::find_binary(const string &name)
{
	SearchPath path;
	if(FS::Path(name).is_absolute())
		path.push_back("/");
	else
	{
		if(sys_bin_path.empty())
		{
			string env_path = Msp::getenv("PATH");
			if(!env_path.empty())
			{
				for(const string &p: split(env_path, ':'))
					sys_bin_path.push_back(p);
			}
			else
			{
				sys_bin_path.push_back("/bin");
				sys_bin_path.push_back("/usr/bin");
			}
		}
		path = sys_bin_path;
	}

	for(const FS::Path &p: path)
	{
		FS::Path filename = p/name;
		if(FileTarget *tgt = get_target(filename))
		{
			builder.get_logger().log("vfs", "Binary %s found in %s as existing %s", name, p, tgt->get_type());
			return tgt;
		}
		else if(file_exists(filename))
		{
			builder.get_logger().log("vfs", "Binary %s found in %s", name, p);
			return new Executable(builder, filename);
		}

		builder.get_logger().log("vfs", "Binary %s not found in %s", name, p);
	}

	return 0;
}

bool VirtualFileSystem::file_exists(const FS::Path &filename)
{
	if(nonexistent.count(filename))
		return false;
	if(FS::is_reg(filename))
		return true;
	nonexistent.insert(filename);
	return false;
}
